#ifndef __cuckoo_block_h__
#define __cuckoo_block_h__

#include "cuckoo-common.h"

template <class K, class HF, class EMPTY>
class CuckooBlock : public CuckooDictionary<K,HF,EMPTY> {
protected:
  int d1;
  HF h[2];


public:
  CuckooBlock( int _d = 4, const double &_epsilon=0.1 ) :
  CuckooDictionary<K,HF,EMPTY>( _d, _epsilon ) {
    this->d1 = _d-1;
    reserve( 4*_d );
  };

  int insert( const K &_x ) {
    K x( _x );
    int freecell = not_an_index;

    int r = binrand();
    int i0 = h[r]( x )*this->d;
    int i1 = h[1-r]( x )*this->d;
    int i = i0;
    int ende = i0+this->d;

    for( ; i0 < ende; i0++, i1++ ) {
      if( this->mem[i0] == x )
        return found;
      else
        if( freecell == not_an_index &&
            this->__empty.is_empty(this->mem[i0]) )
          freecell = i0;
      if( this->mem[i1] == x )
        return found;
      else
        if( freecell == not_an_index &&
            this->__empty.is_empty(this->mem[i1]) )
          freecell = i1;
      if( this->__empty.is_empty(this->mem[i0]) &&
          this->__empty.is_empty(this->mem[i1]) )
        break;
    }

    this->account -= 2;

    if( freecell != not_an_index ) {
      this->mem[freecell] = x;
      this->m_nkeys++;
      return freecell;
    }

    while( this->account > 0 ) {
      int j = i + my_rand( ) % this->d;
      ::swap( x, this->mem[j] );
      r = binrand();
      i0 = h[r]( x )*this->d;
      i = i0;
      int ende = i0+this->d;
      this->account--;
      for( ; i0 < ende; i0++ ) {
        if(this->__empty.is_empty(this->mem[i0]) ){
          this->mem[i0] = x;
          this->m_nkeys++;
          return i0;
        }
      }
    }
    this->remaining_key = x;
    return not_an_index;
  }

  void remove
    ( const K & x ) {
    int i0 = h[0]( x ) * this->d;
    int i1 = h[1]( x ) * this->d;
    int block_ende0 = i0+this->d;
    int block_ende1 = i1+this->d;
    int ende, zelle = not_an_index;
    for( ; i0 < block_ende0; i0++, i1++ ) {
      if( this->mem[i0] == x ) {
        ende = block_ende0;
        zelle = i0;
        break;
      }
      if( this->mem[i1] == x ) {
        ende = block_ende1;
        zelle = i1;
        break;
      }
    }
    if( zelle == not_an_index )
      return;
    int _zelle = zelle;
    zelle++;
    while( zelle < ende &&
           !this->__empty.is_empty(this->mem[zelle]) )
      zelle++;
    zelle--; // Zelle zeigt auf den letzten belegten Index

    if( zelle > _zelle )
      this->mem[_zelle] = this->mem[zelle];
    this->mem[zelle] = this->__empty;
    this->m_nkeys--;
  }

  int insert_grow( const K & _x ) {
    this->account = this->insert_grow_account;
    int index = insert( _x );
    if( index != not_an_index )
      return index;
    if( !grow_by( this->f_grow ) )
      return not_an_index;
    return insert( this->remaining_key );
  };

  void clear( ) {
    for( int j = this->m-1; j >= 0; j-- )
      this->mem[ j ] = this->__empty;
    this->m_nkeys = 0;
    this->remaining_key = this->__empty;
  }

  /** Versucht, ncells Zellen zu reservieren und
  die Schlssel dort unterzubringen. Wenn das 
  nicht geht, wird false zurckgegeben. */
  bool reserve( int ncells ) {
    HF _h0 = h[0], _h1 = h[1];
    int oldm = this->m;
    int oldnkeys = this->m_nkeys;
    K old_remaining_key = this->remaining_key;
    int old_account = this->account;

    int nBlocks = (ncells+this->d1) / this->d;
    h[0] = HF( nBlocks );
    h[1] = HF( nBlocks );
    ncells = nBlocks * this->d;

    K *oldmem = this->mem;
    this->mem = new K[ ncells ];
    this->m = ncells;
    this->set_account( (int)(this->account_factor * this->m) );
    this->clear( );
    this->remaining_key = old_remaining_key;
    if( oldmem != NULL ) {
      for( int j = oldm-1; j >= 0; j-- ) {
        if( !this->__empty.is_empty(oldmem[j]) ) {
          if( insert(oldmem[j]) == not_an_index ) {
            h[0] = _h0;
            h[1] = _h1;
            delete[] this->mem;
            this->mem = oldmem;
            this->m = oldm;
            this->m_nkeys = oldnkeys;
            this->remaining_key = old_remaining_key;
            this->account = old_account;
            return false;
          }
        }
      }
      delete[] oldmem;
    }
    return true;
  };

  bool grow_by( const double &factor ) {
    int _m = (int)ceil( factor * this->m );
    if( factor > 1 && _m == this->m )
      _m++;
    return reserve( _m );
  }

  inline int lookup( const K &x ) const {
    int i0 = h[0]( x ) * this->d;
    if( this->mem[i0] == x )
      return i0;
    int i1 = h[1]( x ) * this->d;
    if( this->mem[i1] == x )
      return i1;
    int ende=i0+this->d;
    i0++;
    i1++;
    for( ; i0 < ende; i0++, i1++ ) {
      if( this->mem[i0] == x )
        return i0;
      if( this->mem[i1] == x )
        return i1;
    }
    return not_an_index;
  }
};

#endif
