#ifndef __cuckoo_dary_h__
#define __cuckoo_dary_h__

#include "cuckoo-common.h"

template <class K, class HF, class EMPTY>
class CuckooDary : public CuckooDictionary<K,HF,EMPTY>{
protected:
  HF *h;
  int d1;

public:
  CuckooDary( int _d = 4, const double &_epsilon=0.1 ):
  CuckooDictionary<K,HF,EMPTY>( _d, _epsilon ){
    d1 = d-1;
    h = new HF[_d];
    reserve( 4*this->d );
  };

  ~CuckooDary( ) {
    delete [] h;
  };

  int insert( const K &_x ) {
    K x( _x );
    int freecell = not_an_index;
    int i = 0;

    int r = my_rand( ) % this->d;
    for( int j = this->d-1; j >= 0; j-- ) {
      i = h[r]( x );
      if( this->mem[i] == x )
        return found;
      else
        if( freecell == not_an_index &&
            this->__empty.is_empty( this->mem[i] ) )
          freecell = i;
      r++;
      if( r == this->d )
        r -= this->d;
    }

    this->account -= this->d;

    if( freecell != not_an_index ) {
      this->mem[freecell] = x;
      this->m_nkeys++;
      return freecell;
    }

    while( --this->account > 0 ) {
      ::swap( x, this->mem[i] );
      if( this->__empty.is_empty(x) ) {
        this->m_nkeys++;
        return i;
      }
      r = my_rand( ) % d1;
      int oldi = i;
      i = h[r]( x );
      if( i == oldi )
        i = h[d1]( x );
    }
    this->remaining_key = x;
    return not_an_index;
  }

  void remove
    ( const K & _x ) {
    int index = lookup( _x );
    if( index != not_an_index ) {
      this->mem[index] = this->__empty;
      this->m_nkeys--;
    }
  }

  int insert_grow( const K & _x ) {
    this->account = this->insert_grow_account;
    int index = insert( _x );
    if( index != not_an_index )
      return index;
    if( !grow_by( this->f_grow ) )
      return not_an_index;
    return insert( this->remaining_key );
  };

  /** Versucht, ncells Zellen zu reservieren
      und die Schlssel dort unterzubringen. 
      Wenn das nicht geht, wird false zurckgegeben. */
  bool reserve( int ncells ) {
    HF *_h = h;
    h = new HF[this->d];
    int oldm = this->m;
    int oldnkeys = this->m_nkeys;
    K old_remaining_key = this->remaining_key;
    int old_account = this->account;
    for( int j = this->d-1; j >= 0; j-- ) {
      h[ j ] = HF( ncells / this->d, j * (ncells/this->d) );
    }
    K *oldmem = this->mem;
    this->mem = new K[ ncells ];
    this->m = ncells;
    this->set_account( (int)(this->account_factor * this->m) );
    this->clear( );
    this->remaining_key = old_remaining_key;
    if( oldmem != NULL ) {
      for( int j = oldm-1; j >= 0; j-- ) {
        if( !this->__empty.is_empty(oldmem[j]) ) {
          if( insert(oldmem[j]) == not_an_index ) {
            delete[] h;
            delete[] this->mem;
            h = _h;
            this->mem = oldmem;
            this->m = oldm;
            this->m_nkeys = oldnkeys;
            this->remaining_key = old_remaining_key;
            this->account = old_account;
            return false;
          }
        }
      }
      delete[] oldmem;
    }
    delete[] _h;
    return true;
  };

  bool grow_by( const double &factor ) {
    int _m = (int)ceil( factor * this->m );
    if( factor > 1 && _m == this->m )
      _m++;
    return reserve( _m );
  }

  inline int lookup( const K &x ) const {
    for( int j = this->d-1; j >= 0; j-- ) {
      int i = h[j]( x );
      if( this->mem[i] == x )
        return i;
    }
    return not_an_index;
  }

  double get_avg_probes( ) const {
    return max( 2*this->d+1., 2/log(1.+this->epsilon));
  };

  double get_element_probes( ) const {
    double res =
      2*std::max(
        this->d+1.,
        ::pow(1/this->epsilon, ::log(2.0*this->d)) );
    return res;
  };
};

#endif
